import json
import arrow
import random
import time
import requests
import logging
from jose.utils import base64url_encode, base64url_decode
from .arweave_lib import Transaction
from .utils import *
from .merkle import validate_path, CHUNK_SIZE
from .arweave_lib import API_URL

from signal import signal, SIGPIPE, SIG_DFL
signal(SIGPIPE, SIG_DFL)

logger = logging.getLogger(__name__)

MAX_CHUNKS_IN_BODY = 1

FATAL_CHUNK_UPLOAD_ERRORS = ['invalid_json', 'chunk_too_big', 'data_path_too_big', 'offset_too_big', 'data_size_too_big', 'chunk_proof_ratio_not_attractive', 'invalid_proof']

ERROR_DELAY = 1000 * 40


class TransactionUploaderException(Exception):
    pass


class TransactionUploader:
    def __init__(self, *args, **kwargs):
        self.chunk_index = kwargs.get('chunk_index', 0)
        self.tx_posted = kwargs.get('tx_posted', False)
        self.transaction = kwargs['transaction']
        self.last_request_time_end = kwargs.get('last_request_time_end', 0)
        self.last_response_status = kwargs.get('last_response_status', 0)
        self.last_response_error = kwargs.get('last_response_error', '')
        self.transaction.data = b''  # zero out data for serialization
        self.file_handler = kwargs['file_handler']
        self.total_errors = 0
        self.data = None

    @property
    def is_complete(self):
        return self.tx_posted and self.chunk_index == len(self.transaction.chunks.get("chunks"))

    @property
    def total_chunks(self):
        return len(self.transaction.chunks.get("chunks"))

    @property
    def uploaded_chunks(self):
        return self.chunk_index

    @property
    def pct_complete(self):
        return int("{}".format(self.uploaded_chunks / self.total_chunks * 100).split('.')[0])

    def to_json(self):
        data = {
            "chunkIndex": self.chunk_index,
            "transaction": self.transaction.to_dict(),
            "lastRequestTimeEnd": self.last_request_time_end,
            "lastResponseStatus": self.last_response_status,
            "lastResponseError": self.last_response_error,
            "lastResponseError": self.tx_posted
        }

        return json.dumps(data)

    def load_from_json(self, data):
        if type(data) == str:
            data = json.loads(data)

        self.chunk_index = data['chunkIndex']
        self.transaction = Transaction.from_serialized_transaction(data['transaction'])
        self.last_request_time_end = arrow.now(data['lastRequestTimeEnd']).timestamp
        self.last_response_status = data['lastResponseStatus']
        self.last_response_error = data['lastResponseError']
        self.tx_posted = data['lastResponseError']

    def upload_chunk(self):
        if self.is_complete:
            raise TransactionUploaderException("Upload is already complete")

        if self.last_response_error != '':
            self.total_errors += 1
        else:
            self.total_errors = 0

        if self.total_errors == 100:
            raise TransactionUploaderException(
                "Unable to complete upload: {}: {}".format(self.last_response_status, self.last_response_error)
            )

        delay = 0

        if self.last_response_error != '':
            delay = max(
                (self.last_request_time_end + ERROR_DELAY) - arrow.now().timestamp,
                ERROR_DELAY
            )

        if delay > 0:
            delay = delay - (delay * random.random() * 0.3)
            time.sleep(delay)

        self.last_response_error = ''

        chunk = self.transaction.get_chunk(self.chunk_index)

        if not self.tx_posted:
            self.post_transaction(chunk)

        if self.is_complete:
            return

        chunk_ok = validate_path(
            self.transaction.chunks.get('data_root'),
            int(chunk.get('offset')),
            0,
            int(chunk.get('data_size')),
            base64url_decode(chunk.get('data_path'))
        )

        if not chunk_ok:
            raise TransactionUploaderException("Unable to validate chunk {}".format(self.chunk_index))

        self.data = chunk['chunk']  # = self.get_chunk_data(self.chunk_index)
        chunk['data_path'] = chunk['data_path'].decode()
        chunk['chunk'] = chunk['chunk'].decode()

        url = "{}/chunk".format(self.transaction.api_url)

        headers = {'Content-Type': 'application/json', 'Accept': 'text/plain'}

        response = requests.post(url, data=json.dumps(chunk), headers=headers)

        # logger.error("{}\n\n{}".format(response.text, self.transaction.json_data))

        if response.status_code == 200:
            logger.debug("RESPONSE 200: {}".format(response.text))
        else:
            logger.error("{}".format(response.text))

            return {"status": -1, "data": {"error": response.text}}

        self.last_request_time_end = arrow.now().timestamp
        self.last_response_status = response.status_code

        if self.last_response_status == 200:
            self.chunk_index += 1
        else:
            self.last_response_error = json.loads(response.text)

            if self.last_response_error.error in FATAL_CHUNK_UPLOAD_ERRORS:
                raise TransactionUploaderException(
                    "Fatal error uploading chunk {}: {}".format(self.chunk_index, self.last_response_error.error)
                )

    def get_chunk_data(self, chunk_index):
        self.file_handler.seek(chunk_index * CHUNK_SIZE)
        data = self.file_handler.read(CHUNK_SIZE)

        return base64url_encode(data)

    def post_transaction(self, chunk):
        upload_in_body = self.total_chunks <= MAX_CHUNKS_IN_BODY

        if upload_in_body:
            url = "{}/tx".format(self.transaction.api_url)
            headers = {'Content-Type': 'application/json', 'Accept': 'text/plain'}

            self.transaction.data = chunk['chunk']

            json_data = self.transaction.json_data
            response = requests.post(url, data=json_data, headers=headers)

            self.last_request_time_end = arrow.now().timestamp
            self.last_response_status = response.status_code
            self.transaction.data = b''

            if 200 <= response.status_code < 300:
                logger.debug("RESPONSE 200: {}".format(response.text))
                self.tx_posted = True
                self.chunk_index = MAX_CHUNKS_IN_BODY
                return
            else:
                logger.error("{}\n\n{}".format(response.text, self.transaction.json_data))

                self.last_response_error = json.loads(response.text)

                raise TransactionUploaderException(
                    "Unable to upload transaction {}, {}".format(response.status_code, self.last_response_error)
                )

        url = "{}/tx".format(self.transaction.api_url)
        headers = {'Content-Type': 'application/json', 'Accept': 'text/plain'}

        self.transaction.data = b''

        json_data = self.transaction.json_data
        response = requests.post(url, data=json_data, headers=headers)

        self.last_request_time_end = arrow.now().timestamp
        self.last_response_status = response.status_code
        self.transaction.data = b''

        if not (200 <= response.status_code < 300):
            self.last_response_error = json.loads(response.text)

            raise TransactionUploaderException(
                "Unable to upload transaction {}, {}".format(response.status_code, self.last_response_error)
            )

        self.tx_posted = True


class TransactionDownloaderException(Exception):
    pass


def get_transaction_offset(tx_id):
    url = "{}/tx/{}/offset".format(API_URL, tx_id)

    response = requests.get(url)
    response_json = json.loads(response.text)

    if response.status_code == 200:
        return int(response_json.get("data"))
    else:
        raise TransactionDownloaderException(
            "Unable to get transaction offset: {}".format(response_json.get("error"))
        )

def get_chunk(offset):
    url = "{}/chunk/{}".format(API_URL, offset)

    response = requests.get(url)
    response_json = json.loads(response.text)

    if response.status_code == 200:
        return response_json.get("data")
    else:
        raise TransactionDownloaderException(
            "Unable to get chunk: {}".format(response_json.get("error"))
        )


def get_chunk_data(offset):
    chunk = get_chunk(offset)
    buf = base64url_decode(chunk.get('chunk'))
    return buf


def first_chunk_offset(offset_response):
    return int(offset_response.get('offset')) - int(offset_response.get('size')) + 1


def download_chunked_data(tx_id, file_handler=None):
    offset_response = get_transaction_offset(tx_id)

    size = int(offset_response.get('size'))
    end_offset = int(offset_response.get('offset'))
    start_offset = end_offset - size + 1

    byte_offset = 0

    if file_handler is None:
        data = bytearray(length=size)

    while start_offset + byte_offset < end_offset:
        chunk_data = get_chunk_data(start_offset + byte_offset)

        if file_handler is None:
            data[byte_offset:] = byte(chunk_data)
        else:
            file_handler.seek(byte_offset)
            file_handler.write(chunk_data)

        



def from_serialized(self, file_handler, json_str):
    if json_str is None:
        raise TransactionUploaderException("Serialized object does not match expected format")

    serialized = json.loads(json_str)

    if type(serialized.chunk_index) != int or type(serialized.transaction) != object:
        raise TransactionUploaderException("Serialized object does not match expected format")

    upload = TransactionUploader(
        file_handler=file_handler,
        transaction=Transaction(
            transaction=serialized.transaction
        )
    )


def from_transaction_id(file_handler, transaction_str, wallet, api_url=API_URL):
    tx = json.loads(transaction_str)

    url = "{}/tx/{}".format(api_url, tx.get('id'))

    headers = {'Content-Type': 'application/json', 'Accept': 'text/plain'}

    response = requests.get(url, headers=headers)

    logger.error("{}".format(response.text))

    if response.status_code == 200:
        logger.debug("RESPONSE 200: {}".format(response.text))
    else:
        raise TransactionUploaderException("Tx {} not found: {} {}".format(
            tx.get('id'),
            response.status_code,
            response.text)
        )

    serialized = TransactionUploader(
        tx_posted=True,
        chunk_index=0,
        last_response_error='',
        last_request_time_end=0,
        last_response_status=0,
        file_handler=file_handler,
        transaction=Transaction(
            wallet,
            transaction=response.text
        )
    )

    return serialized


def get_uploader(upload, file_handler):
    uploader = None
    if type(upload) == Transaction:
        uploader = TransactionUploader(file_handler=file_handler, transaction=upload)
    else:
        if type(upload) == str:
            upload = from_transaction_id(file_handler, upload)

        uploader = TransactionUploader(file_handler=file_handler, transaction=upload)

    return uploader



import hashlib
from Crypto.Signature import PKCS1_PSS
from Crypto.Hash import SHA256
from jose import jwk
from jose.utils import base64url_encode, base64url_decode, base64


def create_tag(name, value, v2):
    if v2:
        b64name = name
        b64value = value
    else:
        b64name = base64url_encode(name.encode('ascii')).decode()
        b64value = base64url_encode(value.encode('ascii')).decode()
    
    return {"name": b64name, "value":b64value}


def encode_tag(tag):
    b64name = base64url_encode(tag['name'].encode('ascii')).decode()
    b64value = base64url_encode(tag['value'].encode('ascii')).decode()

    return {"name": b64name, "value": b64value}

def decode_tag(tag):
    name = base64url_decode(tag['name'].encode())
    value = base64url_decode(tag['value'].encode())
    
    return {'name':name, 'value':value}


def owner_to_address(owner):
    result = base64url_encode(hashlib.sha256(base64url_decode(owner.encode('ascii'))).digest()).decode()

    return result


def winston_to_ar(winston_str):
    length = len(winston_str)
    
    if length > 12:
        past_twelve = length - 12
        winston_str = "{}.{}".format(winston_str[0:past_twelve], winston_str[-12:])
    else:
        lessthan_twelve = 12 - length
        winston_str = "0.{}{}".format(winston_str,"0" * lessthan_twelve)
        
    return float(winston_str)


def ar_to_winston(ar_amount):
    ar_str = "{:.12f}".format(ar_amount)
    return ar_str.replace('.','').lstrip("0")


def concat_buffers(buffers):
    total_length = 0

    for buffer in buffers:
        total_length += len(buffer)

    offset = 0

    temp = b'\x00' * total_length
    temp = bytearray(temp)
    for buffer in buffers:
        for i in range(len(buffer)):
            temp[i + offset] = buffer[i]

        offset += len(buffer)

    return bytes(temp)




import hashlib
from jose.utils import base64url_encode, base64url_decode, base64
from jose.backends.cryptography_backend import CryptographyRSAKey
from Crypto.PublicKey import RSA
from Crypto.Signature import PKCS1_PSS
from Crypto.Hash import SHA384


def deep_hash(data):
    if type(data) == list:
        tag = b"list" + str(len(data)).encode()

        return deep_hash_chunks(data, hashlib.sha384(tag).digest())

    else:
        tag = b"blob" + str(len(data)).encode()

    tagged_hash = hashlib.sha384(tag).digest() + hashlib.sha384(data).digest()

    return hashlib.sha384(tagged_hash).digest()


def deep_hash_chunks(chunks, acc):
    if len(chunks) < 1:
        return acc

    hash_pair = acc + deep_hash(chunks[0])

    new_acc = hashlib.sha384(hash_pair).digest()

    return deep_hash_chunks(chunks[1:], new_acc)


def read_file_chunks(file_handler, chunk_size, seek_to=0):
    """A generator function to read files one chunk at a time"""
    while True:
        data = file_handler.read(chunk_size)

        if not data:
            break

        yield data



import hashlib
import struct
import functools
from jose.utils import base64url_encode, base64url_decode
from .file_io import read_file_chunks
from .utils import concat_buffers
from json import JSONEncoder

CHUNK_SIZE = 256 * 1024
NOTE_SIZE = 32
HASH_SIZE = 32
MAX_CHUNK_SIZE = 256 * 1024
MIN_CHUNK_SIZE = 32 * 1024


class NodeTypeException(Exception):
    pass


class Node:
    def __init__(self, id='', type=None, byte_range=0, max_byte_range=0):
        self.id = id
        self.type = type
        self.byte_range = byte_range
        self.max_byte_range = max_byte_range


class BranchNode(Node):
    def __init__(self, *args, **kwargs):
        super(BranchNode, self).__init__(id=kwargs['id'], max_byte_range=kwargs['max_byte_range'], byte_range=kwargs['byte_range'])
        self.type = "branch"
        self.left_child = kwargs.get("left_child", None)
        self.right_child = kwargs.get("right_child", None)


class LeafNode(Node):
    def __init__(self, *args, **kwargs):
        super(LeafNode, self).__init__(max_byte_range=kwargs['max_byte_range'])
        self.data_hash = kwargs['data_hash']
        self.min_byte_range = kwargs['min_byte_range']
        self.id = hash([hash(self.data_hash), hash(int_to_buffer(self.max_byte_range))])
        self.type = "leaf"


class TaggedChunk:
    def __init__(self, tc_id, end):
        self.id = tc_id
        self.end = end

class Chunk:
    def __init__(self, data_hash, data_size=0, min_byte_range=0, max_byte_range=0):
        self.data_size = data_size
        self.data_hash = data_hash
        self.min_byte_range = min_byte_range
        self.max_byte_range = max_byte_range

    def to_dict(self):
        print("boom!")
        return {
            "dataHash": base64url_encode(self.data_hash).decode(),
            "maxByteRange": self.max_byte_range,
            "minByteRange": self.min_byte_range
        }

class HashNode:
    def __init__(self, hn_id, max):
        self.id = hn_id
        self.max = max

class Proof:
    def __init__(self, offset, proof):
        self.offset = offset
        self.proof = proof

    def to_dict(self):
        return {
            "offset": self.offset,
            "proof": base64url_encode(self.proof).decode()
        }


class ValidatedPathResult:
    def __init__(self, offset, left_bound, right_bound, chunk_size):
        self.offset = offset
        self.left_bound = left_bound
        self.right_bound = right_bound
        self.chunk_size = chunk_size


def chunk_data(file_handler):
    """
    Takes the input data and chunks it into (mostly) equal sized chunks.
    The last chunk will be a bit smaller as it contains the remainder
    from the chunking process.
    :param file_handler:
    :return: chunks
    """
    chunks = []; chadd = chunks.append

    cursor = 0

    for chunk in read_file_chunks(file_handler, MAX_CHUNK_SIZE):
        data_hash = hashlib.sha256(chunk).digest()

        cursor += len(chunk)

        chadd(
            Chunk(
                data_hash,
                data_size=len(chunk),
                min_byte_range=cursor - len(chunk),
                max_byte_range=cursor
            )
        )

    return tuple(chunks)  # lets make this a fast processing tuple for later!


def compute_root_hash(file_handler):
    root_node = generate_tree(file_handler)

    return root_node.id


def generate_leaves(chunks):
    leaves = [
        LeafNode(
            data_hash=chunk.data_hash,
            min_byte_range=chunk.min_byte_range,
            max_byte_range=chunk.max_byte_range
        )
        for chunk in chunks
    ]

    return tuple(leaves)


def generate_tree(file_handler):
    root_node = build_layers(generate_leaves(chunk_data(file_handler)))

    return root_node


def build_layers(nodes, level=0):
    nodes_lenth = len(nodes)

    if nodes_lenth < 2:
        left = nodes[0]
        right = None if nodes_lenth == 1 else nodes[1]
        root = hash_branch(left, right)

        return root

    next_layer = []; nadd = next_layer.append

    for i in range(0, nodes_lenth, 2):
        left = nodes[i]
        right = None if i+1 > (nodes_lenth-1) else nodes[i+1]

        nadd(hash_branch(left, right))

    return build_layers(next_layer, level + 1)


def generate_proofs(root):
    proofs = resolve_branch_proofs(root)

    if type(proofs) == Proof:
        proofs = (proofs,)

    if type(proofs) != tuple:
        return flatten_list(proofs)

    return flatten_tuple(proofs)


def generate_transaction_chunks(file_handler):
    chunks = chunk_data(file_handler)
    leaves = generate_leaves(chunks)
    root = build_layers(leaves)
    proofs = generate_proofs(root)

    last_chunk = chunks[-1]
    if last_chunk.max_byte_range - last_chunk.min_byte_range == 0:
        chunks = chunks[:-1]
        proofs = proofs[:-1]

    return {
        "data_root": root.id,
        "chunks": chunks,
        "proofs": proofs
    }


def flatten_tuple(inputs):
    flat = []; fadd = flat.append

    for item in inputs:
        if type(item) == tuple:
            fadd(flatten_tuple(item))
        else:
            fadd(item)

    return tuple(flat)


def flatten_list(inputs):
    flat = []; fadd = flat.append; fexd = flat.extend

    for item in inputs:
        if type(item) == list:
            fexd(flatten_list(item))
        else:
            fadd(item)

    return flat


def resolve_branch_proofs(node, proof=b'', depth=0):
    if node.type == "leaf":
        return Proof(
            node.max_byte_range - 1,
            concat_buffers([proof, node.data_hash, int_to_buffer(node.max_byte_range)])
        )

    if node.type == "branch":
        partial_proof = concat_buffers([
            proof,
            node.left_child.id,
            node.right_child.id,
            int_to_buffer(node.byte_range)
        ])

        return [
            resolve_branch_proofs(node.left_child, partial_proof, depth + 1),
            resolve_branch_proofs(node.right_child, partial_proof, depth + 1),
        ]

    raise NodeTypeException("Unexpected node type")


def hash_branch(left, right=None):
    if not right:
        return left

    return BranchNode(
        id=hash(
            [
                hash(left.id),
                hash(right.id),
                hash(int_to_buffer(left.max_byte_range))
            ]
        ),
        byte_range=left.max_byte_range,
        max_byte_range=right.max_byte_range,
        left_child=left,
        right_child=right
    )


def hash_leaf(data, note):
    return HashNode(
        hash([hash(data), hash(note_to_buffer(note))]),
        note
    )


def hash(data):
    if type(data) == list:
        byte_str = b''
        for line in data:
            byte_str += line

        data = byte_str

    digest = hashlib.sha256(data).digest()
    b64_str = base64url_encode(digest)
    return digest


def note_to_buffer(note):
    buffer = b"\x00" * NOTE_SIZE
    buffer = bytearray(buffer)

    for i in range(NOTE_SIZE-1, 0, -1):
        if i > 0:
            if note > 0:
                buffer[i] = note.to_bytes(4, byteorder='big')[-1]
                note = note >> 8
            else:
                break
        else:
            break

    return bytes(buffer)


def int_to_buffer(note):
    buffer = b"\x00" * NOTE_SIZE
    buffer = bytearray(buffer)

    for i in range(NOTE_SIZE - 1, 0, -1):
        byte_val = note % 256
        buffer[i] = int(byte_val)
        note = int((note - byte_val) / 256)

    return buffer


def buffer_to_int(buffer):
    value = 0

    for byte_val in buffer:
        value *= 256
        value += byte_val

    return value


def array_compare(a, b):
    functools.reduce(lambda x, y: x and y, map(lambda p, q: p == q, a, b), True)


def validate_path(id, dest, left_bound, right_bound, path):
    if right_bound < 0:
        return False

    if dest > right_bound:
        return validate_path(id, 0, right_bound - 1, right_bound, path)

    if dest < 0:
        return validate_path(id, 0, 0, right_bound, path)

    if len(path) == HASH_SIZE + NOTE_SIZE:
        path_data = path[0:HASH_SIZE]
        path_data_length = len(path_data)
        end_offset_buffer = path[path_data_length:path_data_length + NOTE_SIZE]

        path_data_hash = hash([
            hash(path_data),
            hash(end_offset_buffer)
        ])

        result = id == path_data_hash

        if result:
            return ValidatedPathResult(right_bound-1, left_bound, right_bound, right_bound - left_bound)

        return False

    left = path[:HASH_SIZE]
    left_length = len(left)
    right = path[left_length: left_length + HASH_SIZE]
    right_length = len(right)

    offset_buffer = path[left_length + right_length: left_length + right_length + NOTE_SIZE]
    offset = buffer_to_int(offset_buffer)

    remainder = path[left_length + right_length + len(offset_buffer):]

    path_hash = hash([
        hash(left),
        hash(right),
        hash(offset_buffer)
    ])

    if id == path_hash:
        if dest < offset:
            return validate_path(
                left,
                dest,
                left_bound,
                min(right_bound, offset),
                remainder
            )

        return validate_path(
            right,
            dest,
            max(left_bound, offset),
            right_bound,
            remainder
        )

    return False

def debug(proof, output=""):
    if len(proof) < 1:
        return output

    left = proof[:HASH_SIZE]
    right = proof[len(left):len(left) + HASH_SIZE]
    offset_buffer = proof[
        len(left) + len(right) : len(left) + len(right) + HASH_SIZE
    ]

    offset = buffer_to_int(offset_buffer)

    remainder = proof[:len(left) + len(right) + len(offset_buffer)]

    path_hash = hash([
        hash(left),
        hash(right),
        hash(offset_buffer)
    ])

    updated_output = "{}\n{},{},{} => {}".format(
        output,
        bytearray(left),
        bytearray(right),
        offset,
        bytearray(path_hash)
    )

    return debug(remainder, updated_output)